from itertools import zip_longest, chain
import textwrap
import copy

from genbank.sequence import Seq

def rev_comp(dna):
	a = 'acgtrykmbvdh'
	b = 'tgcayrmkvbhd'
	tab = str.maketrans(a,b)
	return dna.translate(tab)[::-1]

def grouper(iterable, n, fillvalue=None):
	args = [iter(iterable)] * n
	return zip_longest(*args, fillvalue=fillvalue)

def nint(s):
	return int(s.replace('<','').replace('>',''))


class Feature():
	def __init__(self, type_, strand, pairs, locus, tags=None):
		#super().__init__(locus.locus, locus.dna)
		self.type = type_
		self.strand = strand
		# tuplize the pairs
		self.pairs = tuple([tuple(pair) for pair in pairs])
		self.locus = locus
		self.tags = tags if tags else dict()
		#self.dna = ''
		#self.partial = False

	def frame(self, end='left'):
		if self.type != 'CDS':
			return 0
		elif end == 'right' or self.partial() == 'left':
			return (self.right()%3+1) * self.strand
		elif end == 'left' or self.partial() == 'right':
			return (self.left()%3+1) * self.strand

	def left(self):
		# convert genbank 1-based indexing to standard 0-based
		# this is probably not right for <1 that should -2
		return nint(self.pairs[0][0]) - 1
	
	def right(self):
		# convert genbank 1-based indexing to standard 0-based
		return nint(self.pairs[-1][-1]) - 3
	
	def length(self):
		return len(self.seq())

	'''
	def seq(self):
		seq = list()
		for pair in self.pairs:
			left,right = map(nint, pair)
			seq.append(self.locus.seq(left-1,right, self.strand))
		if self.strand > 0:
			return ''.join(seq)
		else:
			return ''.join(seq[::-1])
	'''

	def seq(self):
		# lol this is chatgpts more optimized version
		seq = [self.locus.seq(nint(pair[0]) - 1, nint(pair[-1]), self.strand) for pair in self.pairs]
		return ''.join(seq) if self.strand > 0 else ''.join(seq[::-1])

	'''
	def loc(self):
		loc = list()
		for pair in self.pairs:
			left,right = map(nint, pair)
			loc.extend(list(range(left-1,right)))
		return loc
	'''
	def loc(self):
		# lol this is chatgpts more optimized version
		return [val for pair in self.pairs for val in range(nint(pair[0])-1, nint(pair[-1]))]
	
	def codons(self, loc=False):
		assert self.type == 'CDS'
		dna = self.seq()
		# should I return partial codons?
		partial_type = self.partial()
		remainder = len(dna) % 3
		if partial_type == 'left':
			dna = dna[remainder:] if self.strand > 0 else dna[:-remainder]
		elif partial_type == 'right':
			dna = dna[:-remainder] if self.strand > 0 else dna[remainder:]
		if len(dna) % 3:
			# not sure how to handle bad features
			dna = ''
		for triplet in grouper(dna, 3):
			yield ''.join(triplet)
		return

	def codon_locations(self):
		assert self.type == 'CDS'
		loc = self.loc()
		if self.strand < 0:
			loc.reverse()
		# should I return partial codons?
		partial_type = self.partial()
		remainder = len(loc) % 3
		if partial_type == 'left':
			loc = loc[remainder:] if self.strand > 0 else loc[:-remainder]
		elif partial_type == 'right':
			loc = loc[:-remainder] if self.strand > 0 else loc[remainder:]
		if len(loc) % 3:
			# not sure how to handle bad features
			loc = []
		for locs in grouper(loc, 3):
			yield locs
		return


	def fna(self):
		return self.header() + self.seq() + "\n"

	def faa(self):
		return self.header() + self.translation() + "\n"
	
	def header(self):
		header = ">" + self.locus.name() + "_CDS_[" + self.locations() + "]"
		for tag,values in self.tags.items():
			if tag != 'translation':
				for value in values:
					if value:
						header += " [" + tag + "=" + value +"]"
					else:
						header += " [" + tag +"]"
		return header + "\n"

	def hypothetical(self):
		function = ''.join(self.tags['product']).lower() if 'product' in self.tags else ''
		if 'hypot'  in function or \
		   'etical' in function or \
		   'unchar' in function or \
		   ('orf' in function and 'orfb' not in function):
			return True
		else:
			return False

	def partial(self):
		partial_left  = any(['<' in item for pair in self.pairs for item in pair])
		partial_right = any(['>' in item for pair in self.pairs for item in pair])
		if partial_left and partial_right:
			# this really shouldnt even happen, maybe raise an error?
			return 'both'
		elif partial_left:
			return 'left'
		elif partial_right:
			return 'right'
		else:
			return None

	def is_type(self, _type):
		if self.type == _type:
			return True
		else:
			return False

	def is_joined(self):
		if len(self.pairs) > 1:
			return True
		return False

	'''
	def __iter__(self):
		for left,*right in self.pairs:
			if right:
				right = right[0]
			else:
				right = left
			yield nint(left)-1 , nint(right)-1
	'''
	def __str__(self):
		"""Compute the string representation of the feature."""
		return "%s\t%s\t%s\t%s" % (
				repr(self.locus.name()),
				repr(self.type),
				repr(self.pairs),
				repr(self.tags))

	def __repr__(self):
		"""Compute the string representation of the feature."""
		return "%s(%s, %s, %s, %s)" % (
				self.__class__.__name__,
				repr(self.locus.name()),
				repr(self.type),
				repr(self.pairs),
				repr(self.tags))
	def __hash__(self):
		return hash(self.pairs)
	#def __eq__(self, other):
	#	return self.pairs == other.pairs()

	def __lt__(self, other):
		if self.left() == other.left():
			return self.right() < other.right()
		else:
			return self.left() < other.left()

	def locations(self):
		pairs = []
		#for left, *right in self.pairs:
		for pair in self.pairs:
			pairs.append("..".join(pair))
		location = ','.join(pairs)
		if len(pairs) > 1:
			location = 'join(' + location + ')'
		if self.strand < 0:
			location = 'complement(' + location + ')'
		return location


	def split(self):
		a = copy.copy(self)
		b = copy.copy(self)
		return a,b

	def translation(self):
		aa = []
		codon = ''
		first = self.length() % 3 if (self.partial() == 'left' and self.strand > 0) or (self.partial() == 'right' and self.strand < 0) else 0
		dna = self.seq()
		for i in range(first, self.length(), 3):
			codon = dna[ i : i+3 ]
			aa.append(self.locus.translate.codon(codon))
		#if self.strand < 0:
		#	aa = aa[::-1]
		# keeping the stop codon character adds 'information' as does which of
		# the stop codons it is. It is the better way to write the fasta
		#if aa[-1] in '#*+':
		#	aa.pop()
		# keeping the first amino acid also adds 'information' to the fasta
		#aa[0] = 'M'
		return "".join(aa)

	def write(self, outfile):
		outfile.write('     ')
		outfile.write( self.type.ljust(16) )
		if self.strand < 0:
			outfile.write('complement(')
		# the pairs
		if len(self.pairs) > 1:
			outfile.write('join(')
		pairs = []
		#for left, *right in self.pairs:
		for pair in self.pairs:
			#pair = left + '..' + str(right[0]) if right else str(left)
			pairs.append("..".join(pair))
		outfile.write(','.join(pairs))
		if len(self.pairs) > 1:
			outfile.write(')')
		# the pairs
		if self.strand < 0:
			outfile.write(')')
		outfile.write('\n')
		for tag,values in self.tags.items():
			for value in values:
				if value is not None:
					for line in textwrap.wrap( '/' + str(tag) + '=' + str(value) , 58):
						outfile.write('                     ')
						outfile.write(line)
						outfile.write('\n')
				else:
					outfile.write('                     ')
					outfile.write('/' + str(tag))
					outfile.write('\n')




def rev_comp(dna):
	a = 'acgtrykmbvdh'
	b = 'tgcayrmkvbhd'
	tab = str.maketrans(a,b)
	return dna.translate(tab)[::-1]


class Seq(str):
	# this is just to capture negative string indices as zero
	def __getitem__(self, key):
		if isinstance(key, slice):
			if key.stop and key.stop < 0:
				key = slice(0, 0, key.step)
			elif key.start and key.start < 0:
				key = slice(0, key.stop, key.step)
		elif isinstance(key, int) and key >= len(self):
			return ''
		return super().__getitem__(key)



import io
import sys
import gzip
import tempfile
from itertools import chain as chain

from genbank.locus import Locus


class File(dict):
	formats = ['genbank','gff', 'gff3','tabular',
			   'fasta', 'fna','faa','bases',
			   'coverage', 'rarity', 'gc', 'gcfp', 'taxonomy','part',
			   'testcode']
	def __init__(self, filename=None):
		if not hasattr(self, 'locus'):
			self.locus = Locus
		''' use tempfiles since using next inside a for loop is easier'''
		temp = tempfile.TemporaryFile()
		
		lib = gzip if filename.endswith(".gz") else io
		with lib.open(filename, mode="rb") as fp:
			for line in chain(fp, [b'>']):
				# FASTA
				if line.startswith(b'>') and temp.tell() > 1:
					locus = self.parse_locus(temp)
					self[locus] = True
				temp.write(line)
				# GENBANK
				if line.startswith(b'//'):
					locus = self.parse_locus(temp)
					self[locus] = True 
		temp.close()
	
	def __init_subclass__(cls, locus, **kwargs):
		'''this method allows for a Locus class to be modified through inheritance in other code '''
		super().__init_subclass__(**kwargs)
		cls.locus = locus

	'''
	def __iter__(self):
		# hopefully this doesnt break stuff
		return iter(self.values())
	'''

	def features(self, include=None, exclude=None):
		for locus in self:
			for feature in locus.features(include=include):
				yield feature
	
	def dna(self):
		dna = [locus.dna for locus in self.values()]
		return "".join(dna)

	def parse_locus(self, fp):
		locus = self.locus()
		current = fasta = False
		dna = []	
		fp.seek(0)
		for line in fp:
			line = line.decode("utf-8")
			if line.startswith('\n'):
				pass
			elif not line.startswith(' '):
				# ITS A NEW MAIN GROUP
				group,*value = line.split(maxsplit=1)
				if line.startswith('>'):
					locus.groups['LOCUS'] = [group[1:]]
					fasta = True
				elif fasta:
					dna.append( line.rstrip().replace(' ','').lower() )
				elif line.startswith('//'):
					break
				else:
					locus.groups.setdefault(group, []).append(''.join(map(str, value)))
					#if group in locus.groups and locus.groups[group][-1]:
					#	locus.groups[group].append(''.join(map(str,value)))
					#else:
					#	locus.groups[group] = [''.join(map(str,value))]
			elif group == 'ORIGIN':
				#locus.dna += line.split(maxsplit=1)[1].rstrip().replace(' ','').lower()
				dna.append( line.split(maxsplit=1)[1].rstrip().replace(' ','').lower() )
			elif group == 'FEATURES':
				line = line.rstrip()
				if not line.startswith(' ' * 21):
					while line.endswith(','):
						line += next(fp).decode('utf-8').strip()
					current = locus.read_feature(line)
				else:
					while line.count('"') == 1:
						line += " " + next(fp).decode('utf-8').strip()
					tag,sep,value = line[22:].partition('=')
					#current.tags[tag] = value #.replace('"', '')
					if sep:
						current.tags.setdefault(tag, []).append(value)
					else:
						current.tags.setdefault(tag, []).append(None)
			else:
				locus.groups[group][-1] += line

		locus.dna = ''.join(dna)
		fp.seek(0)
		if fp.writable():
			fp.truncate()
		return locus

	def write(self, outfile=sys.stdout):
		for locus in self:
			locus.write(outfile)




import re
import sys
import textwrap
from collections.abc import Sequence
from itertools import chain

from genbank.codons import Last
from genbank.codons import Next
from genbank.codons import Codons
from genbank.feature import Feature
from genbank.feature import grouper
from genbank.sequence import Seq
from genbank.translate import Translate

def rev_comp(dna):
	a = 'acgtrykmbvdh'
	b = 'tgcayrmkvbhd'
	tab = str.maketrans(a,b)
	return dna.translate(tab)[::-1]


def nint(s):
	return int(s.replace('<','').replace('>',''))

def rmap(func, items):
	out = list()
	for item in items:
		if isinstance(item, Sequence) and not isinstance(item, str):
			out.append(type(item)(rmap(func,item)))
		else:
			out.append(func(item))
	return type(items)(out)

def recursive_map(func, items):
    return (recursive_map(func, x) if isinstance(x, tuple) else func(x) for x in items)


class Locus(dict):
	def __init__(self, name='', dna=''):
		if not hasattr(self, 'feature'):
			self.feature = Feature
		#self.name = name
		self.dna = dna.lower()
		#self.codons = dict()
		self.translate = Translate()
		self.strand = 1
		self.groups = dict()
		self.groups['LOCUS'] = [name.replace(' ','')] if name else []
		#self.groups['FEATURES'] = ['']
		#self.groups['ORIGIN'] = ['']
		
	def __eq__(self, other):
		return (self == other)

	def __hash__(self):
		return id(self)

	def __init_subclass__(cls, feature=Feature, **kwargs):
		'''this method allows for a Feature class to be modified through inheritance in other code '''
		super().__init_subclass__(**kwargs)
		cls.feature = feature

	def name(self):
		return self.groups['LOCUS'][0].split(' ')[0]

	def molecule(self):
		if len(locus) > 2:
			return locus[3]
		else:
			return 'DNA'

	def fasta(self):
		return ">" + self.name() + "\n" + self.seq() + "\n"

	def seq(self, left=0, right=None, strand=None):
		# this should always refer to zero based indexing
		#if strand is None:
		#	strand = self.strand
		if left < 0:
			left = None
		if right and right < 0:
			right = 0
		if strand and strand < 0:
			return Seq(rev_comp(self.dna[left : right]))
		else:
			return Seq(         self.dna[left : right] )

	def length(self):
		return len(self.dna)

	def gc_content(self, seq=None):
		if seq is not None:
			#a = seq.count('a') + seq.count('A')
			c = seq.count('c') + seq.count('C')
			g = seq.count('g') + seq.count('G')
			#t = seq.count('t') + seq.count('T')
			return (c+g) / len(seq) #(a+c+g+t)
		elif not hasattr(self, "gc"):
			#a = self.dna.count('a') + self.dna.count('A')
			c = self.dna.count('c') + self.dna.count('C')
			g = self.dna.count('g') + self.dna.count('G')
			#t = self.dna.count('t') + self.dna.count('T')
			self.gc = (c+g) / len(self.dna) # (a+c+g+t)
		return self.gc

	def pcodon(self, codon):
		codon = codon.lower()
		seq = self.dna + rev_comp(self.dna)
		p = dict()
		p['a'] = seq.count('a') / len(seq)
		p['c'] = seq.count('c') / len(seq)
		p['g'] = seq.count('g') / len(seq)
		p['t'] = seq.count('t') / len(seq)
		return p[codon[0]] * p[codon[1]] * p[codon[2]]

	def rbs(self):
		for feature in self:
			if feature.type == 'CDS':
				if feature.strand > 0:
					start = feature.left()+3
					feature.tags['rbs'] = self.seq(start-30,start)
				else:
					start = feature.right()
					feature.tags['rbs'] = rev_comp(self.seq(start,start+30))
	
	def features(self, include=None, exclude=None):
		for feature in self:
			if not include or feature.type in include:
				yield feature

	def add_feature(self, key, strand, pairs, tags=dict()):
		"""Add a feature to the factory."""
		#feature = self.feature
		strand = 44 - ord(strand) if strand in ['+','-'] else 0 if strand in ['.'] else int(strand)
		feature = self.feature(key, strand, pairs, self, tags=tags)
		if feature not in self:
			self[feature] = len(self)
		return feature

	def read_feature(self, line):
		"""Add a feature to the factory."""
		key = line.split()[0]
		val = line.split()[1]
		#partial  = 'left' if '<' in line else ('right' if '>' in line else False)
		strand = -1 if 'complement' in line else 1
		# this is for weird malformed features
		if ',1)' in line:
			line = line.replace( ",1)" , ",1..1)" )
		#pairs = [pair.split('..') for pair in re.findall(r"<*\d+\.\.>*\d+", line)]
		#pairs = [map(int, pair.split('..')) for pair in re.findall(r"<?\d+\.{0,2}>?\d+", line.replace('<','').replace('>','') )]
		#pairs = [ pair.split('..') for pair in re.findall(r"<?\d+\.{0,2}>?[-0-9]*", line) ]
		pairs = [ pair.split('..') for pair in re.findall(r"<?\d+\.{0,2}>?\d*", val) ]
		# tuplize the pairs
		pairs = tuple([tuple(pair) for pair in pairs])
		feature = self.add_feature(key, strand, pairs)
		return feature

	def gene_coverage(self):
		''' This calculates the protein coding gene coverage, which should be around 1 '''
		cbases = tbases = 0	
		index = [ [False] * len(self.dna) , [False] * len(self.dna)]
		for feature in self.features(include=['CDS','tRNA']):
			for locations in grouper(feature.loc(), 3):
				for location in locations:
					if location:
						index[(feature.strand >> 1) * -1][location-1] = True
					break
		cbases += sum([item for sublist in index for item in sublist])
		tbases += len(self.dna) / 3
		return cbases , tbases
	
	def gc_fp(self):
		fp = [0,0,0]
		for feature in self.features(include=['CDS']):
			for codon in feature.codons():
				for i,base in enumerate(codon):
					fp[i] += (ord(base) >> 1 ) & 3 % 2
		return fp

	def testcode(self):
		pos = [
				{ 'a': .22, 'c': .23, 'g': .08, 't': .09},
				{ 'a': .20, 'c': .30, 'g': .08, 't': .09},
				{ 'a': .34, 'c': .33, 'g': .16, 't': .20},
				{ 'a': .45, 'c': .51, 'g': .27, 't': .54},
				{ 'a': .68, 'c': .48, 'g': .48, 't': .44},
				{ 'a': .58, 'c': .66, 'g': .53, 't': .69},
				{ 'a': .93, 'c': .81, 'g': .64, 't': .68},
				{ 'a': .84, 'c': .70, 'g': .74, 't': .91},
				{ 'a': .68, 'c': .70, 'g': .88, 't': .97},
				{ 'a': .94, 'c': .80, 'g': .90, 't': .97}
				]
		con = [
				{ 'a': .21, 'c': .31, 'g': .29, 't': .58},
				{ 'a': .81, 'c': .39, 'g': .33, 't': .51},
				{ 'a': .65, 'c': .44, 'g': .41, 't': .69},
				{ 'a': .67, 'c': .43, 'g': .41, 't': .56},
				{ 'a': .49, 'c': .59, 'g': .73, 't': .75},
				{ 'a': .62, 'c': .59, 'g': .64, 't': .55},
				{ 'a': .55, 'c': .64, 'g': .64, 't': .40},
				{ 'a': .44, 'c': .51, 'g': .47, 't': .39},
				{ 'a': .49, 'c': .64, 'g': .54, 't': .24},
				{ 'a': .28, 'c': .82, 'g': .40, 't': .28}
				]
		wp = {'a':.26,'c':.18,'g':.31,'t':.33}
		wc = {'a':.11,'c':.12,'g':.15,'t':.14}
		content = {
		'a':[0,0,0],
		'c':[0,0,0],
		'g':[0,0,0],
		't':[0,0,0]
		}
		for i,base in enumerate(self.seq()):
			content[base][i%3] += 1
		position = dict()
		for base in 'acgt':
			position[base] = max(content[base]) / (min(content[base])+1)
		p = 0
		for base in 'acgt':
			n = int(str(position[base])[2]) if position[base] < 2 else 9
			p += pos[n][base] * wp[base]
		for base in 'acgt':
			percent = sum(content[base]) / self.length()
			n = int(str((percent-0.17) * 5 )[2])+1 if percent > 0.17 else 0
			p += con[n][base] * wc[base]
		if p <= 0.75:
			return 'noncoding'
		elif p < 0.95:
			return 'no opinion'
		else:
			return 'coding'



	def write(self, outfile=sys.stdout, args=None):
		if not args or args.format == 'genbank':
			self.write_gbk(outfile)
		elif args.format == 'gff':
			self.write_gff(outfile)
		elif args.format == 'gff3':
			self.write_gff3(outfile)

	def write_gbk(self, outfile=sys.stdout):
		for group,values in chain(self.groups.items(), [[None,[True, False]]] ):
			for value in values:
				if group == 'LOCUS':
					outfile.write('LOCUS       ')
					cols = self.groups['LOCUS'][0].split(' ')
					# I eventually need to properly format the locus line
					outfile.write(self.name().ljust(9))
					outfile.write(str(len(self.dna)).rjust(19))
					outfile.write(' bp ')
					if 'bp' in cols:
						outfile.write(' '.join(cols[cols.index('bp')+1:]))
					else:
						outfile.write('\n')
					continue
				elif group == 'FEATURES' or (not group and value and 'FEATURES' not in self.groups):
					outfile.write('FEATURES             Location/Qualifiers\n')
					for feature in self:
						feature.write(outfile)
				elif group == 'ORIGIN' or (not group and not value and 'ORIGIN' not in self.groups):
					# should there be spaces after ORIGIN?
					outfile.write('ORIGIN      ')
					i = 0
					dna = textwrap.wrap(self.dna, 10)
					for block in dna:
						if(i%60 == 0):
							outfile.write('\n')
							outfile.write(str(i+1).rjust(9))
							outfile.write(' ')
							outfile.write(block.lower())
						else:
							outfile.write(' ')
							outfile.write(block.lower())
						i += 10
				elif group == 'BASE':
					for value in values:
						outfile.write(group)
						outfile.write(' ')
						outfile.write(value)
				elif group:
					outfile.write(group.ljust(12))
					outfile.write(value)
		outfile.write('\n')
		outfile.write('//')
		outfile.write('\n')

	def write_gff3(self, outfile=sys.stdout):
		outfile.write('>Feature')
		outfile.write(' ') # should this be a space or a tab?
		outfile.write(self.name())
		outfile.write('\n')
		outfile.write('1')
		outfile.write('\t')
		outfile.write(str(self.length()))
		outfile.write('\t')
		outfile.write('REFERENCE')
		outfile.write('\n')
		for feature in self.features(include=['CDS']):
			pairs = [list(item)[::feature.strand] for item in feature.pairs][::feature.strand]
			pair = pairs.pop(0)
			outfile.write(pair[0])
			outfile.write("\t")
			outfile.write(pair[-1])
			outfile.write("\t")
			outfile.write(feature.type)
			for pair in pairs:
				outfile.write("\n")
				outfile.write(pair[0])
				outfile.write("\t")
				outfile.write(pair[-1])
			for tag,values in feature.tags.items():
				for value in values:
					outfile.write("\n")
					outfile.write("\t\t\t")
					outfile.write(str(tag))
					outfile.write("\t")
					if value is None:
						pass
					elif value[0] == '"' and value[-1] == '"':
						outfile.write(value[1:-1])
					else:
						outfile.write(value)
			outfile.write("\n")

	def write_gff(self, outfile=sys.stdout):
		outfile.write("##gff-version 3\n")
		for feature in self:
			outfile.write(self.name())
			outfile.write("\t")
			if 'SOURCE' in self.groups:
				outfile.write(self.groups['SOURCE'][0].split('\n')[0])
			else:
				outfile.write('None')
			outfile.write("\t")
			outfile.write(feature.type)
			outfile.write("\t")
			outfile.write(feature.pairs[0][0])
			outfile.write("\t")
			outfile.write(feature.pairs[-1][-1])
			outfile.write("\t")
			if hasattr(feature, 'score'):
				outfile.write(str(feature.score))
			else:
				outfile.write('.')
			outfile.write("\t")
			outfile.write(str(feature.strand).replace('-1','-').replace('1','+'))
			outfile.write("\t")
			outfile.write(".")
			outfile.write("\t")
			for tag,values in feature.tags.items():
				for value in values:
					outfile.write(str(tag))
					if value is None:
						pass
					elif value[0] == '"' and value[-1] == '"':
						outfile.write("=")
						outfile.write(value[1:-1])
					else:
						outfile.write("=")
						outfile.write(value)
				outfile.write(";")
			outfile.write("\n")


	def last(self, n, codons, strand):
		# this needs to be 0-based indexing
		if isinstance(codons, str):
			codons = [codons.lower()]
		codons = [codon.lower() for codon in codons]
		if strand > 0:
			irange = range(n,            -1, -3)
		else:
			irange = range(n, self.length(), +3)

		for i in irange:
			if self.seq(i,i+3,strand) in codons:
				return i
		return None

	def next(self, n, codons, strand):
		if isinstance(codons, str):
			codons = [codons]
		codons = [codon.lower() for codon in codons]
		if strand > 0:
			irange = range(n, self.length(), +3)
		else:
			irange = range(n,            -1, -3)
		for i in irange:
			if self.seq(i,i+3,strand) in codons:
				return i
		return None

	def nearest(self, n, codons, strand):
		_last = self.last(n,strand,codons)
		if not _last:
			_last = 0
		_next = self.next(n,strand,codons)
		if not _next:
			_next = self.length()
		if n - _last < _next - n:
			return _last
		else:
			return _next

	def distance(self, n, strand, codons):
		nearest = self.nearest(n, strand, codons)
		return n - nearest if nearest < n else nearest - n

	def codon_rarity(self, codon=None):
		if not hasattr(self, 'rarity'):
			seen = {-1:dict(), +1:dict()}
			self.rarity = {a+b+c : 0 for a in 'acgt' for b in 'acgt' for c in 'acgt'}
			for feature in self:
				if feature.type == 'CDS':
					#for _codon, _loc in zip(feature.codons(), feature.codon_locations()):
					#	if _codon in self.rarity and _loc not in seen[feature.strand]:
					for _codon in feature.codons():
						if _codon in self.rarity:
							self.rarity[_codon] += 1
					#		seen[feature.strand][_loc] = True
		total = sum(self.rarity.values())
		self.rarity = {codon:self.rarity[codon]/total for codon in self.rarity}
		if codon in self.rarity:
			return self.rarity[codon]
		elif codon:
			return None
		else:
			return self.rarity

	def slice(self, left, right):
		if left > right:
			left,right = right+1,left+1
			self.strand = -1
		self.dna = self.seq(left,right)
		to_delete = list()
		for feature in self.keys():
			if feature.right() - 1 < left or feature.left() > right:
				to_delete.append(feature)
			else:
				# whew there is a lot going on here
				f0 = lambda x : int(x.replace('<','').replace('>',''))
				f1 = lambda x : '<1' if f0(x) - left < 1 else ('>'+str(self.length()) if f0(x) - left > self.length() else f0(x) - left)
				f2 = lambda x : str(f1(x))
				feature.pairs = rmap(f2, feature.pairs)
		for feature in to_delete:
			del self[feature]
		return self






#!/usr/bin/env python3
from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE,SIG_DFL) 
import os
import io
import sys
import gzip
import re
import argparse
import tempfile
from collections import Counter
from itertools import zip_longest, chain
import shutil
import tempfile
import urllib.request
import fileinput

sys.path.pop(0)
from genbank.file import File

def get(x):
	return True

def is_valid_file(x):
	if not os.path.exists(x):
		raise argparse.ArgumentTypeError("{0} does not exist".format(x))
	return x

def nint(x):
    return int(x.replace('<','').replace('>',''))

def _print(self, item):
    if isinstance(item, str):
        self.write(item)
    else:
        self.write(str(item))

if __name__ == "__main__":
	usage = '%s [-opt1, [-opt2, ...]] infile' % __file__
	parser = argparse.ArgumentParser(description='', formatter_class=argparse.RawTextHelpFormatter, usage=usage)
	parser.add_argument('infile', type=is_valid_file, help='input file in genbank format')
	parser.add_argument('-o', '--outfile', action="store", default=sys.stdout, type=argparse.FileType('w'), help='where to write output [stdout]')
	parser.add_argument('-f', '--format', help='Output the features in the specified format', type=str, default='genbank', choices=File.formats)
	parser.add_argument('-s', '--slice', help='This slices the infile at the specified coordinates. \nThe range can be in one of three different formats:\n    -s 0-99      (zero based string indexing)\n    -s 1..100    (one based GenBank indexing)\n    -s 50:+10    (an index and size of slice)', type=str, default=None)
	parser.add_argument('-g', '--get', action="store_true")
	parser.add_argument('-r', '--revcomp', action="store_true")
	parser.add_argument('-a', '--add', help='This adds features the shell input via < features.txt', type=str, default=None)
	parser.add_argument('-e', '--edit', help='This edits the given feature key with the value from the shell input via < new_keys.txt', type=str, default=None)
	parser.add_argument('-k', '--key', help='Print the given keys [and qualifiers]', type=str, default=None)
	parser.add_argument('-c', '--compare', help='Compares the CDS of two genbank files', type=str, default=None)
	args = parser.parse_args()
	args.outfile.print = _print.__get__(args.outfile)

	if not args.get:
		genbank = File(args.infile)
	else:
		#raise Exception("not implemented yet")
		# not ready yet
		accession,rettype = args.infile.split('.')
		url = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nuccore&id=' + accession + '&rettype=' + rettype + '&retmode=text'
		with urllib.request.urlopen(url) as response:
			with tempfile.NamedTemporaryFile() as tmp:
				shutil.copyfileobj(response, tmp)
				genbank = File(tmp.name)
		
	if args.compare:
		perfect = partial = total = fp = 0
		compare = File(args.compare)
		partial = dict()
		perfect = dict()
		mistake = dict()
		for locus,other in zip(genbank,compare):
			pairs = dict()
			for feature in locus.features(include='CDS'):
				if feature.strand > 0:
					pairs[feature.pairs[-1][-1]] = feature.pairs[ 0][ 0]
				else:
					pairs[feature.pairs[ 0][ 0]] = feature.pairs[-1][-1]
			total += len(pairs)
			for feature in other.features(include='CDS'):
				start,stop = feature.pairs[ 0][ 0],feature.pairs[-1][-1]
				if feature.strand < 0:
					stop,start = start,stop
				if stop in pairs:
					partial[stop] = True
					if start == pairs[stop]:
						perfect[stop] = True
				else:
					mistake[stop] = True
					#print(feature)
		partial = len(partial)
		perfect = len(perfect)
		mistake = len(mistake)
		args.outfile.write(f"{partial}\t{partial/total}\t{perfect}\t{perfect/total}\t{total}\t{mistake}\n")
		exit()
	if args.add:
		# this only works for single sequence files
		stdin = []
		if not sys.stdin.isatty():
			stdin = sys.stdin.readlines()
		for locus in genbank:
			for line in stdin:
				if args.add == 'genbank':
					pass
				elif args.add == 'genemarks':
					if line.startswith(' ') and 'Gene' not in line and '#' not in line:
						n,strand,left,right,*_ = line.split()
						locus.add_feature('CDS',strand,[[left,right]],{'note':['genemarkS']})
				elif args.add == 'genemark':
					if line.startswith(' ') and (' direct ' in line or ' complement ' in line):
						left,right,strand,*_ = line.split()
						strand = '+' if ('direct' in line or '+' in line) else '-'
						locus.add_feature('CDS',strand,[[left,right]],{'note':['genemark']})
					elif line.startswith('List of Regions of interest'):
						break
				elif args.add == 'glimmer':
					if not line.startswith('>'):
						_,left,right,(strand,*_),*_ = line.split()
						if strand == '-':
							left, right = right, left
						if int(left) > int(right):
							if strand == '-':
								pairs = [['1', right], [left, str(locus.length())]]
							else:
								pairs = [[left, str(locus.length())], ['1', right]]
						else:
							pairs = [[left,right]]
						locus.add_feature('CDS',strand,pairs,{'note':['glimmer3']})
				elif args.add == 'gff':
					if not line.startswith('#') and len(line) > 2:
						try:
							name,other,key,left,right,_,strand,_,tags,*_ = line.rstrip('\n').split('\t')
						except:
							print('Error:')
							print(line)
							exit()
						key = key[:14]
						tags = {key: ['"%s"' % '='.join(val)] for tag in tags.split(';') for key,*val in [tag.split('=')]}
						locus.add_feature(key,strand,[[left,right]],tags)
				else:
					if not line.startswith('#') and len(line) > 2:
						mapping = dict()
						for key,val in zip(args.add.split(','),line.split()):
							mapping[key] = val
						left,right,strand = [mapping[key] for key in ['left','right','strand']]
						locus.add_feature('CDS',strand,[[left,right]],{})
							

	elif args.edit:
		if not sys.stdin.isatty():
			stdin = sys.stdin.readlines()
			#sys.stdin = open('/dev/tty')
		key,qualifier = args.edit.replace('/',':').split(':')
		for feature,values in zip(genbank.features(include=[key]), stdin):
			feature.tags[qualifier] = list()
			for value in values.rstrip().split('\t'):
				feature.tags[qualifier].append(value)
	if args.slice:
		if '..' in args.slice:
			left,right = map(int, args.slice.split('..'))
			left -= 1
		elif ':' in args.slice:
			left,right = args.slice.split(':')
			if '+' in right and '-' in right:
				left = eval(left + right)
				right = eval(left + right)
			elif '+' in right:
				right = eval(left + right)
			elif '-' in right:
				left,right = eval(left + right) , left
			left,right = map(int, [left,right])
		elif '-' in args.slice:
			left,right = map(int, args.slice.split('-'))
			right += 1
		else:
			raise Exception("re-circularization not implemented yet")
			left = int(args.slice)
			right = left+1
		for locus in genbank:
			locus = locus.slice(left,right)
	if args.key:
		key,qualifier = args.key.replace('/',':').split(':')
		for feature in genbank.features(include=key):
			args.outfile.print('\t'.join(feature.tags[qualifier]))
			args.outfile.print("\n")
	elif args.format == 'genbank':
		if args.revcomp:
			raise Exception("not implemented yet")
		genbank.write(args.outfile)	
	elif args.format == 'tabular':
		for feature in genbank.features(include=['CDS']):
			args.outfile.print(feature)
			args.outfile.print("\t")
			args.outfile.print(feature.seq())
			args.outfile.print("\n")
	elif args.format in ['gff', 'gff3']:
		for locus in genbank:
			locus.write(args.outfile, args)
	elif args.format in ['fna','faa']:
		for locus in genbank:
			for feature in locus.features(include=['CDS']):
				args.outfile.print( getattr(feature, args.format)() )
	elif args.format in ['fasta']:
		for locus in genbank:
			if args.revcomp:
				locus.dna = locus.seq(strand=-1)
			args.outfile.print( getattr(locus, args.format)() )
	elif args.format == 'coverage':
		cbases = tbases = 0
		for locus in genbank:
			c,t = locus.gene_coverage()
			cbases += c
			tbases += t
		#args.outfile.print( name )
		#args.outfile.print( '\t' )
		args.outfile.print( cbases / tbases )
		args.outfile.print( '\n' )
	elif args.format == 'rarity':
		rarity = dict()
		for locus in genbank:
			for codon,freq in sorted(locus.codon_rarity().items(), key=lambda item: item[1]):
				args.outfile.print(codon)
				args.outfile.print('\t')
				args.outfile.print(round(freq,5))
				args.outfile.print('\n')
	elif args.format == 'bases':
		strand = -1 if args.revcomp else +1
		for locus in genbank:
			args.outfile.print(locus.seq(strand=strand))
			args.outfile.print('\n')
	elif args.format in ['gc','gcfp']:
		for locus in genbank:
			args.outfile.print(locus.name())
			args.outfile.print('\t')
			if args.format == 'gc':
				args.outfile.print(locus.gc_content())
			else:
				args.outfile.print(locus.gc_fp())
			args.outfile.print('\n')
	elif args.format == 'taxonomy':
		for locus in genbank:
			args.outfile.print(locus.groups['SOURCE'][0].replace('\n','\t').replace('            ','').replace(';\t','; ') )
			args.outfile.print('\n')
	elif args.format in ['part']:
		folder = args.outfile.name if args.outfile.name != '<stdout>' else ''
		for locus in genbank:
			with open(os.path.join(folder,name + '.fna'), 'w') as f:
				f.write('>')
				f.write(locus.name())
				f.write('\n')
				f.write(locus.seq())
				f.write('\n')
	elif args.format == 'testcode':
		for locus in genbank:
			args.outfile.print(locus.name())
			args.outfile.print('\t')
			args.outfile.print(locus.testcode())
			args.outfile.print('\n')




